"""
PaSh: Light-touch Data-Parallel Shell Processing

A system for parallelizing POSIX shell scripts.
Hosted by the Linux Foundation.

Usage:
    pash script.sh              # Run a script with PaSh
    pash -c "echo hi | cat"     # Run a command with PaSh
    pash --help                 # Show help

For more information, visit: https://github.com/binpash/pash
"""

__version__ = "0.13.0"
__author__ = "The PaSh Authors"
__license__ = "MIT"
